use std::collections::HashMap;
use serde::{Serialize,Deserialize};

use crate::pkgs::{PkgInfo};
use crate::defs::{TskSrvcName,IsCritical,AppName,Cntrllr};
// use crate::cmds::ssh;
// use tempfile::tempfile;

#[derive(Clone,Debug,Serialize,Deserialize,Default)]
pub struct ConfigTskSrvc {
  pub name: Option<String>,
  pub files: HashMap<String,String>,
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct Provider {
  pub name: String,
  pub runner: String,
  pub args: String,
}
impl Provider {
  pub fn new(name: String, runner: String, args: String) -> Self {
    Self {
      name,
      runner,
      args,
    }
  }
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct CloudEnv {
  pub path: String,
  pub home: String,
  pub monitor_run: String,
  pub config_path: String,
  pub config_json_path: String,
  pub config_root: String,
  pub root_tsksrvcs: String,
  pub provision: String,
  pub cloud: String,
  pub group: String,
  pub target: String,
  pub source: String,
  pub source_path: String,
  pub tpls_path: String,
  pub tsksrvcs_path: String,
  pub wk_path: String,
  pub pkgs_list: String,
  pub versions: String,
  pub listhosts: String,
  pub force: u8,
  key: String,
}

impl CloudEnv {
  pub fn new(force: u8, key: String) -> Self {
    Self {
      path: String::from(""),
      home: String::from(""),
      monitor_run: String::from(""),
      config_root: String::from(""),
      config_path: String::from(""),
      config_json_path: String::from(""),
      root_tsksrvcs: String::from(""),
      tsksrvcs_path: String::from(""),
      provision: String::from(""),
      cloud: String::from(""),
      group: String::from(""),
      target: String::from(""),
      source: String::from(""),
      source_path: String::from(""),
      tpls_path: String::from(""),
      wk_path: String::from(""),
      pkgs_list: String::from(""),
      versions: String::from(""),
      listhosts: String::from(""),
      force,
      key, 
    }
  }
  pub fn get_key(&self) -> String {
    self.key.to_owned()
  }
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct Cluster {
  pub role: String,
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct SSH {
  pub host: String,
  pub keyPath: String,
  pub user: String,
  pub password: String,
  pub port: u16,
}
#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct Networks {
  pub access: String,
  pub family: String,
  pub network: Option<String>,
  pub ipaddress: Option<String>,
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct StorageDevice {
  pub action: String,
  pub title: String,
  pub storage: String,
  pub size: u16,
  pub finalSsize: u16,
  pub partSizes: Option<String>,
  pub makefs: Option<String>,
  pub tier: String,
}
#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct LoginUser {
  pub sshKeys: Vec<String>,
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct Server {
  pub hostname: String,
  pub title: String,
//  pub tplPkgs: Option<String>,
  pub cluster: Cluster,
  pub ssh: SSH,
  pub tags: Vec<String>,
  pub networks: Vec<Networks>,
  pub useFloatIP: bool,
  pub loginUser: LoginUser,
  pub plan: String,
  pub metadata: String,
  pub userData: Option<String>,
  pub timeZone: String,
  pub zone: String,
  pub storageDevices: Vec<StorageDevice>,
  pub tsksrvcs: Vec<TskSrvc>,
//  pub packages: Option<HashMap<String,HashMap<String,serde_yaml::Value>>>,
}

#[allow(non_snake_case)]
pub struct Dhall {
  pub templatedLicense: String,
}

// #[allow(clippy::missing_docs_in_private_items)]
// #[allow(non_snake_case)]
// #[derive(Clone, Debug, Serialize,  Deserialize, Default)]
// pub struct Server {
//   pub name: String,
//   pub tpl: Option<String>,
//   pub provider: Option<String>,
//   pub arch: Option<String>,
//   pub spec: Option<Spec>,
// }

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct CfgRs {
  pub floatIP: Option<String>,
  pub hostPrefix: String,
  pub mainName: String,
  pub provider: Option<String>,
  pub domainName: Option<String>,
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct ConfigResources {
  pub floatIP: Option<String>,
  pub hostPrefix: String,
  pub mainName: String,
  pub provider: Option<String>,
  pub domainName: Option<String>,
	pub cntrllrs: Vec<Cntrllr>,
  pub servers: Vec<Server>,
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct DeployConfig {
  pub floatIP: Option<String>,
  pub hostPrefix: String,
  pub mainName: String,
  pub domainName: Option<String>,
  pub Servers: Vec<Server>,
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct TskSrvc {
  pub name: TskSrvcName,
  pub path: String,
  pub req: String,
  pub target: String,
  pub liveness: String,
	pub critical: IsCritical,
}

#[allow(clippy::missing_docs_in_private_items)]
#[allow(non_snake_case)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct App {
  pub name: AppName,
  pub path: String,
  pub req: String,
  pub target: String,
  pub liveness: String,
  pub critical: IsCritical,
}

// #[allow(clippy::missing_docs_in_private_items)]
// #[derive(Clone, Debug, Serialize,  Deserialize, Default)]
// #[allow(non_snake_case)]
// pub struct TskSrvcs {
//   pub grpName: String,
//   pub tsksrvcsList: Vec<TskSrvc>,
//   pub state: String,
//   pub config: String,
//   pub log: Vec<String>,
// }

#[allow(clippy::missing_docs_in_private_items)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
#[allow(non_snake_case)]
pub struct MainResourcesConfig {
  pub mainName: String,
  pub provider: String,
  pub domainName: String,
} 

#[allow(clippy::missing_docs_in_private_items)]
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
#[allow(non_snake_case)]
pub struct Cloud {
  pub env: CloudEnv,
  pub providers: HashMap<String,Provider>,
  pub state: String,
  pub config_resources: ConfigResources,
  pub full_config: DeployConfig,
  pub tpl: Vec<Server>,
  pub pkgs: HashMap<String, PkgInfo>,
  pub log: Vec<String>,
}
impl Cloud {
  pub async fn load_providers() -> HashMap<String,Provider> {
    let mut providers: HashMap<String,Provider> = HashMap::new();
    providers.insert( "upcloud".to_string(),
      Provider::new("upcloud".to_string(),"upclapi".to_string(),"-kdr tc -c".to_string())
    );
    providers.insert( "manual".to_string(),
      Provider::new("manual".to_string(),"".to_string(),"".to_string())
    );
    providers
  }
}
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct TsksrvcInfo {
	pub name: String,
	pub info: serde_yaml::Value,
}
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct AppsrvcInfo {
	pub name: String,
	pub target: String,
	pub info: serde_yaml::Value,
}
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct SrvcsHostInfo {
	pub hostname: String,
	pub tsksrvcs: Vec<TsksrvcInfo>,
	pub appsrvcs: Vec<AppsrvcInfo>,
}
#[derive(Clone, Debug, Serialize,  Deserialize, Default)]
pub struct HostInfo {
	pub hostname: String,
	pub info: String,
}