use std::collections::HashMap;
use rfm::mkdir;
use anyhow::{anyhow,Result,Context};
use std::fs; //, io};
// use std::os::unix::fs::PermissionsExt;
use std::fs::OpenOptions;
use std::io::{Write};
use std::path::Path;
use flate2::Compression;
use flate2::write::GzEncoder;
use std::process::{Command};

use crate::pkgs::{load_pack_config};
use crate::providers::{upcloud, manual};
// use crate::defs::TskSrvc;
// use crate::cmds::ssh;
// use tempfile::tempfile;

use crate::utils::{parse_yaml_value, write_log, liveness_check}; // host_ssh_is_alive

use crate::defs::{SSHAccess,Cntrllr};
use crate::clouds::defs::{
  ConfigTskSrvc,
  Provider,
  TskSrvc,
  MainResourcesConfig,
  Cloud
};

pub async fn create_tsksrvc_files(cloud: &Cloud, tsk: &TskSrvc, files: HashMap<String,String>, run_pack: &str,tsksrvc_vars: &HashMap<String,String>,tsksrvc_settings: &HashMap<String,serde_yaml::Value>) -> Result<()> {
  let mut tera = tera::Tera::default();
  let mut context = tera::Context::new();
  for (key, value) in tsksrvc_vars {
    if key == "srvr" {
      let srvr_data:serde_yaml::Value = serde_yaml::from_str(&value)?;
      parse_yaml_value(srvr_data,key.to_owned(),&mut context);
	 } else {
      context.insert(key.to_owned(),&value);
    }
  }
  for (key, value) in tsksrvc_settings {
    parse_yaml_value(value.to_owned(),key.to_owned(),&mut context);
  }
  // dbg!(&context);

  for (template_name, value) in files {
    if template_name.starts_with("#") {
      continue;
    }
    let mut output_root_path=format!("{}/{}",&cloud.env.provision,&tsk.name);
    if template_name == run_pack {
      output_root_path = format!("{}",&cloud.env.wk_path);
    }
    let template_path=format!("{}/{}/{}",&cloud.env.tsksrvcs_path,&tsk.name,&template_name);
    let output_path=format!("{}/{}",&output_root_path,template_name.replace(".j2",""));
    // if !Path::new(&output_root_path).exists() {
    //   fs::create_dir(&output_root_path)?;
    // }
    // dbg!(&template_path);
    // dbg!(&output_path);
    let out_path = Path::new(&output_path);
    if let Some(dir_path) = out_path.parent() {
      if !dir_path.exists() {
        let dir_path_buf = dir_path.to_path_buf();
        let dirs: Vec<&std::path::PathBuf> = vec![&dir_path_buf];
        mkdir(&dirs) 
          .with_context(|| format!("\nFailed to create dir path {}", &output_path))?;
        println!("{} created path for ", &output_path);
        // fs::create_dir(&dir_path)?;
      }
    }
    print!(" {} ",&template_name);
    // if value.to_string().contains("tpl") {
    match value.as_str() {
      "tpl" => {
        let template = fs::read_to_string(&template_path)
          .with_context(|| format!("\nFailed to read template from {}", &template_path))?;
        let append = value.to_string().contains("append");
        tera.add_raw_templates(vec![(template_name.to_owned(), &template)])
          .with_context(|| format!("Failed to add template {}", &template_name))?;  
        // dbg!(&context);
        let out=tera
          .render(&template_name, &context)
          .with_context(|| format!("\nFailed to render {}", &template_path))?;  
        //println!("{}",&output_path);
        if append && Path::new(&output_path).exists() {
          let mut file = OpenOptions::new().append(true).open(&output_path)?;
          // out = out.replace("&#x2F;", "/");
          file.write(out.as_bytes())?;
        } else {
          if Path::new(&output_path).exists() {
            fs::remove_file(&output_path)?;
          }
          let mut file = OpenOptions::new().write(true).create(true).open(&output_path)
           .with_context(|| format!("Failed to write template {}", &output_path))?;
          file.write_all(out.as_bytes())?;
          if output_path.ends_with(".sh") {
            // println!("\nSetting permissions for {}",&output_path);
            // let file_sh = fs::File::open(&output_path)?;
            // let metadata = file_sh.metadata()?;
            // let mut permissions = metadata.permissions();
            // permissions.set_mode(0o755);
            Command::new("chmod")
            .arg("+x")
            .arg(&output_path)
            .output()?;
          }
        }
      },
      "cp" => { 
        fs::copy(&template_path, &output_path)?;
      },
      _ => { print!("?"); },
    }
  };
  println!("");
  Ok(())
}
/// ```

pub async fn run_cmd_shell(cmd_name: &str,cloud: &Cloud, cmd: &str, hostname: &str, target_path: String, tsk_name: &str) -> Result<()> {
	// println!("{} {} ..",&cmd_name,&cloud.env.source_path);
	// println!("{} {} ..",&tsk_name,&target_path);
	let output = Command::new(&target_path)
    .arg(&cloud.env.source)
		.arg(&tsk_name)
		.arg(&cmd)
		.arg(&hostname)
		.output()?;
	// dbg!(&output);
	if !&output.status.success() {
		return Err(anyhow!("run '{}' failed: {}",&target_path,&output.status));
	}
	println!("{} {}: {}",&cmd_name,&cloud.env.source_path,&output.status);
  write_log(format!("{}/log",&cloud.env.provision).as_str(), &hostname, &tsk_name, format!("{} {} done",&cmd_name,&target_path).as_str()).await?;
  Ok(())
}
pub async fn run_pack_shell(cloud: &Cloud, tsk: &TskSrvc, cmd: &str, hostname: &str, tsk_name: &str, run_pack_path: String) -> Result<()> {
	let output_root_path=format!("{}/{}",&cloud.env.provision,&tsk.name);
	let archive_file=format!("{}/{}_{}.tar.gz",&cloud.env.wk_path,&hostname,&tsk.name);
	let tar_gz = fs::File::create(&archive_file)?;
	let enc = GzEncoder::new(tar_gz, Compression::default());
	let mut tar = tar::Builder::new(enc);
	tar.append_dir_all(&tsk_name, &output_root_path)?;
	tar.into_inner()?;
	let output = Command::new("bash")
		.arg(&run_pack_path)
		.arg(&cmd)
		.output()?;
	// dbg!(&output);
	if !&output.status.success() {
		return Err(anyhow!("run '{}' on '{}' failed: {}",&run_pack_path,&tsk.name,&output.status));
	} 
	println!("run {}: {}",&tsk.name,&output.status);
	write_log(format!("{}/log",&cloud.env.provision).as_str(), &hostname, &tsk_name, "run").await?;
  Ok(())
}
pub fn tsksrvc_vars_into_settings(root_path: &str, source_path: String, tsk_name: &str, settings: &mut HashMap<String,serde_yaml::Value>) -> Result<()> { 
	let tsksrvc_data= fs::read_to_string(&source_path)
		.with_context(|| format!("Failed to read 'tsksrvcs_path' from {}", &source_path))?;  
	let tsksrvc_settings: HashMap<String,serde_yaml::Value> = serde_yaml::from_str(&tsksrvc_data)?;
	for (key, value) in tsksrvc_settings {
		match key.as_str() {
			"include" => 
				if let Some(val) = value.as_str() {
					for req_tsksrvc in val.split(" ") {
						let include_tsksrvc_path=format!("{}/{}.yaml",&root_path,&req_tsksrvc);
						if Path::new(&include_tsksrvc_path).exists() {
							tsksrvc_vars_into_settings(root_path,include_tsksrvc_path,tsk_name,settings)?;
						}
          }
				}
			,
			"pause" => { run_pause(tsk_name, settings); },
			_ => { settings.insert(key, value); }
		};
	}
  Ok(())
}
pub fn run_pause(tsk_name: &str, settings: &mut HashMap<String,serde_yaml::Value>) {
	let mut num_time: u64 = 8; // as default
	if let Some(time_to_sleep) = settings.get("sleep_time") {
			if let Some(val) = time_to_sleep.as_u64() {
				num_time = val;
			}
	}
	let sleep_time = std::time::Duration::new(num_time, 0);
	println!("Waiting {} tsksrvc for {} seconds ...",&tsk_name,&num_time);
	std::thread::sleep(sleep_time);
}
/// `run_taks` prepare and run a __tsksrvc__ in cloud configuration in a __hostname__ target 
/// 
/// 1) Create a HashMap with __pack_config__ settings for __tsk.name__
/// 2) Create __provision path__ for target cloud 
/// 3) Load __config.yaml__ for target __pack__ (__tsk.name__) 
/// 4) Create a `Tera` default and context with pack and env settings
/// 5) for (template_name, value) in konfig parse target 
/// 6)    Render template if value is 'tpl' of copy file if is 'cp'
/// 7) Pack all files into a '.tar.gz' 
/// 8) run tsksrvc_`tsk.name`_run.sh ( to scp '.tar.gz', run 'install.sh' and clean tsksrvc files
/// 
/// # Examples
///
/// ```no_run
/// ```
pub async fn run_tsksrvc(cloud: &Cloud, tsk: &TskSrvc, cmd: &str, hostname: &str, domain_name: &str, srvr: String, ssh_access: SSHAccess, cntrllrs: &Vec<Cntrllr>) -> Result<()> {
  // load tsksrvc status 
  // if "done" return 
  // host has connection ? 
  let tsk_name = format!("{}",&tsk.name);
	let serverstring = format!("{}:{}",&ssh_access.host,&ssh_access.port);
  liveness_check(&cloud.env.source, &cntrllrs , &serverstring, "", &tsk_name).await?;
  // host_ssh_is_alive(cloud, cmd, hostname, &tsk_name, &ssh_access).await?;
  let prepare_source_path = format!("{}/prepare.sh",&cloud.env.source_path);
  if Path::new(&prepare_source_path).exists() {
		run_cmd_shell("prepare",cloud, cmd, hostname, prepare_source_path, &tsk_name).await?;
  }
  let prepare_path = format!("{}/{}/prepare.sh",&cloud.env.tsksrvcs_path,&tsk.name);
  if Path::new(&prepare_path).exists() {
		run_cmd_shell("prepare",cloud, cmd, hostname, prepare_path, &tsk_name).await?;
  }
  // dbg!(&cloud);
  let mut tsksrvc_vars: HashMap<String,String> = HashMap::new();
  tsksrvc_vars.insert("source".to_string(), cloud.env.source.to_owned());
  tsksrvc_vars.insert("hostname".to_string(), hostname.to_owned());
  tsksrvc_vars.insert("domainName".to_string(), domain_name.to_owned());
  tsksrvc_vars.insert("command".to_string(), cmd.to_owned());
  tsksrvc_vars.insert("srvr".to_string(), srvr.to_owned());
  tsksrvc_vars.insert("tsksrvcname".to_string(), tsk_name.to_owned());
  load_pack_config(&mut tsksrvc_vars, cloud, &tsk, &hostname)?;
  // dbg!(&tsksrvc_vars);

  // create path 
  if !Path::new(&cloud.env.provision).exists() {
    fs::create_dir(&cloud.env.provision)?;
  }
  let run_pack = "run.sh.j2";

  // tsksrvc vars 
  let mut tsksrvc_settings: HashMap<String,serde_yaml::Value> = HashMap::new();
  let tsksrvc_root_path=format!("{}/tsksrvcs",&cloud.env.source_path);

  let global_tsksrvc_path=format!("{}/global.yaml",&tsksrvc_root_path);
  if Path::new(&global_tsksrvc_path).exists() {
		tsksrvc_vars_into_settings(&tsksrvc_root_path,global_tsksrvc_path,&tsk_name,&mut tsksrvc_settings)?;
  }
  let tsksrvc_path=format!("{}/{}.yaml",&tsksrvc_root_path,&tsk.name);
  if Path::new(&tsksrvc_path).exists() {
		tsksrvc_vars_into_settings(&tsksrvc_root_path,tsksrvc_path,&tsk_name,&mut tsksrvc_settings)?;
  }
  let host_tsksrvc_path=format!("{}/{}_{}.yaml",&tsksrvc_root_path,&hostname,&tsk.name);
  if Path::new(&host_tsksrvc_path).exists() {
		tsksrvc_vars_into_settings(&tsksrvc_root_path,host_tsksrvc_path,&tsk_name,&mut tsksrvc_settings)?;
  }
	// dbg!(&tsksrvc_settings);
	if tsk_name.as_str() == "pause" {
		run_pause(&tsk_name, &mut tsksrvc_settings);
	}
  // tsksrvc file config list:
  let tsksrvc_config_path=format!("{}/{}/config.yaml",&cloud.env.tsksrvcs_path,&tsk.name);
  if Path::new(&tsksrvc_config_path).exists() {
    let tsksrvc_config_data= fs::read_to_string(&tsksrvc_config_path)
        .with_context(|| format!("Failed to read 'tsksrvcs_path' from {}", &tsksrvc_config_path))?;  
    let tsksrvc_config: ConfigTskSrvc = serde_yaml::from_str(&tsksrvc_config_data)?;
    create_tsksrvc_files(cloud, tsk, tsksrvc_config.files, &run_pack, &tsksrvc_vars, &tsksrvc_settings).await?;
  }
  // local cloud tsksrvcs config list:
  let local_tsksrvc_config_path=format!("{}/tsksrvcs/config_{}.yaml",&cloud.env.source_path,&tsk.name);
  if Path::new(&local_tsksrvc_config_path).exists() {
    let local_tsksrvc_config_data= fs::read_to_string(&local_tsksrvc_config_path)
        .with_context(|| format!("Failed to read 'tsksrvcs_path' from {}", &local_tsksrvc_config_path))?;  
    let local_tsksrvc_config: ConfigTskSrvc = serde_yaml::from_str(&local_tsksrvc_config_data)?;
    create_tsksrvc_files(cloud, tsk, local_tsksrvc_config.files, &run_pack, &tsksrvc_vars, &tsksrvc_settings).await?;
  }
	// tsksrvc run_pack 
  let run_pack_path=format!("{}/{}",&cloud.env.wk_path,run_pack.replace(".j2",""));
  if Path::new(&run_pack_path).exists() {
    run_pack_shell(cloud, tsk, cmd, hostname, &tsk_name, run_pack_path).await?;
  }
	// tsksrvc postinstall 
	let postinstall_path = format!("{}/{}/postinstall.sh",&cloud.env.tsksrvcs_path,&tsk.name);
	if Path::new(&postinstall_path).exists() {
    run_cmd_shell("postinstall",cloud, cmd, hostname, postinstall_path, &tsk_name).await?;
	}
	// source postinstall 
  let postinstall_source_path = format!("{}/postinstall.sh",&cloud.env.source_path);
  if Path::new(&postinstall_source_path).exists() {
		run_cmd_shell("postinstall",cloud, cmd, hostname, postinstall_source_path, &tsk_name).await?;
  }
  write_log(format!("{}/log",&cloud.env.provision).as_str(), &hostname, &tsk_name, "done").await?;
  Ok(())
}

pub async fn run_tsksrvcs_on_providers(provider: &Provider, cfg_data: &str, cloud: &Cloud, tsksrvc: &str, cmd: &str, nxt: &str, cfg: &MainResourcesConfig ) -> Result<()> {
  match provider.name.as_str() {
    "upcloud" =>  {
      upcloud::run_tsksrvcs_list(cfg_data, cloud, tsksrvc, cmd, &nxt, cfg).await?; 
      Ok(())
    },
    "manual" => {
      manual::run_tsksrvcs_list(cfg_data, cloud, tsksrvc, cmd, &nxt, cfg).await?;
      println!("Provider '{}' create manually",&provider.name);
      Ok(())
    },
    _ => 
      Err(anyhow!("Provider '{}' undefined",&provider.name)),
  }
} 